@@ -17,6 +17,32 @@ use crate::change_set::{
 use crate::{ChangeSetStatus, ResourceStatus, StackEvent, StackEventDetails, StackStatus, Status};
 
 /// The input for the `apply` operation.
+///
+/// You can create an apply input via the [`new`] associated function. Setters are also available to
+/// make construction as ergonomic as possible.
+///
+/// ```no_run
+/// use rusoto_cloudformation::Tag;
+/// # use rusoto_cloudformation::CloudFormationClient;
+/// # use rusoto_core::Region;
+/// use cloudformatious::{ApplyInput, Capability, CloudFormatious, Parameter, TemplateSource};
+///
+/// # #[tokio::main]
+/// # async fn main() -> Result<(), Box<dyn std::error::Error>> {
+/// # let client = CloudFormationClient::new(Region::EuWest2);
+/// let input = ApplyInput::new("my-stack", TemplateSource::inline("{}"))
+///     .set_capabilities([Capability::Iam])
+///     .set_client_request_token("hello")
+///     .set_notification_arns(["arn:foo"])
+///     .set_parameters([Parameter { key: "hello".to_string(), value: "world".to_string() }])
+///     .set_resource_types(["AWS::IAM::Role"])
+///     .set_role_arn("arn:foo")
+///     .set_tags([Tag { key: "hello".to_string(), value: "world".to_string() }]);
+/// let output = client.apply(input).await?;
+/// // ...
+/// # Ok(())
+/// # }
+/// ```
 #[derive(Clone, Debug)]
 #[allow(clippy::module_name_repetitions)]
 pub struct ApplyInput {
@@ -98,32 +124,105 @@ pub struct ApplyInput {
     /// maximum number of 50 tags can be specified.
     pub tags: Vec<Tag>,
 
-    /// Structure containing the template body with a minimum length of 1 byte and a maximum length
-    /// of 51,200 bytes.
+    /// Source for the template body to apply.
     ///
-    /// For more information, go to [Template Anatomy][1] in the AWS CloudFormation User Guide.
-    ///
-    /// Conditional: You must specify either the `template_body` or the `template_url` parameter,
-    /// but not both.
+    /// For more information about templates, go to [Template Anatomy][1] in the AWS CloudFormation
+    /// User Guide.
     ///
     /// [1]: https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/template-anatomy.html
-    pub template_body: Option<String>,
-
-    /// Location of file containing the template body.
-    ///
-    /// The URL must point to a template (max size: 460,800 bytes) that is located in an Amazon S3
-    /// bucket. For more information, go to the [Template Anatomy][1] in the AWS CloudFormation User
-    /// Guide.
-    ///
-    /// Conditional: You must specify either the `template_body` or the `template_url` parameter,
-    /// but not both.
-    ///
-    /// [1]: https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/template-anatomy.html
-    pub template_url: Option<String>,
+    pub template_source: TemplateSource,
 }
 
 impl ApplyInput {
+    /// Construct an input for the given `stack_name` and `template_source`.
+    pub fn new(stack_name: impl Into<String>, template_source: TemplateSource) -> Self {
+        Self {
+            stack_name: stack_name.into(),
+            template_source,
+
+            capabilities: Vec::new(),
+            client_request_token: None,
+            notification_arns: Vec::new(),
+            parameters: Vec::new(),
+            resource_types: None,
+            role_arn: None,
+            tags: Vec::new(),
+        }
+    }
+
+    /// Set the value for `capabilities`.
+    ///
+    /// **Note:** this consumes and returns `self` for chaining.
+    pub fn set_capabilities(mut self, capabilities: impl Into<Vec<Capability>>) -> Self {
+        self.capabilities = capabilities.into();
+        self
+    }
+
+    /// Set the value for `client_request_token`.
+    ///
+    /// **Note:** this consumes and returns `self` for chaining.
+    pub fn set_client_request_token(mut self, client_request_token: impl Into<String>) -> Self {
+        self.client_request_token = Some(client_request_token.into());
+        self
+    }
+
+    /// Set the value for `notification_arns`.
+    ///
+    /// **Note:** this consumes and returns `self` for chaining.
+    pub fn set_notification_arns<I, S>(mut self, notification_arns: I) -> Self
+    where
+        I: Into<Vec<S>>,
+        S: Into<String>,
+    {
+        self.notification_arns = notification_arns
+            .into()
+            .into_iter()
+            .map(Into::into)
+            .collect();
+        self
+    }
+
+    /// Set the value for `parameters`.
+    ///
+    /// **Note:** this consumes and returns `self` for chaining.
+    pub fn set_parameters(mut self, parameters: impl Into<Vec<Parameter>>) -> Self {
+        self.parameters = parameters.into();
+        self
+    }
+
+    /// Set the value for `resource_types`.
+    ///
+    /// **Note:** this consumes and returns `self` for chaining.
+    pub fn set_resource_types<I, S>(mut self, resource_types: I) -> Self
+    where
+        I: Into<Vec<S>>,
+        S: Into<String>,
+    {
+        self.resource_types = Some(resource_types.into().into_iter().map(Into::into).collect());
+        self
+    }
+
+    /// Set the value for `role_arn`.
+    ///
+    /// **Note:** this consumes and returns `self` for chaining.
+    pub fn set_role_arn(mut self, role_arn: impl Into<String>) -> Self {
+        self.role_arn = Some(role_arn.into());
+        self
+    }
+
+    /// Set the value for `tags`.
+    ///
+    /// **Note:** this consumes and returns `self` for chaining.
+    pub fn set_tags(mut self, tags: impl Into<Vec<Tag>>) -> Self {
+        self.tags = tags.into();
+        self
+    }
+
     fn into_raw(self) -> CreateChangeSetInput {
+        let (template_body, template_url) = match self.template_source {
+            TemplateSource::Inline { body } => (Some(body), None),
+            TemplateSource::S3 { url } => (None, Some(url)),
+        };
         CreateChangeSetInput {
             capabilities: Some(self.capabilities.iter().map(ToString::to_string).collect()),
             change_set_name: format!("apply-{}", Utc::now().timestamp_millis()),
@@ -139,8 +238,8 @@ impl ApplyInput {
             role_arn: self.role_arn,
             stack_name: self.stack_name,
             tags: Some(self.tags),
-            template_body: self.template_body,
-            template_url: self.template_url,
+            template_body,
+            template_url,
             ..CreateChangeSetInput::default()
         }
     }
@@ -244,6 +343,48 @@ impl Parameter {
     }
 }
 
+/// Source for a template body.
+///
+/// Templates can be specified for CloudFormation APIs in one of two ways:
+///
+/// - As a JSON string, inline with the request.
+/// - As a URL to a template file on S3.
+///
+/// See the variant documentation for more information.
+#[derive(Clone, Debug)]
+pub enum TemplateSource {
+    /// Structure containing the template body with a minimum length of 1 byte and a maximum length
+    /// of 51,200 bytes.
+    ///
+    /// For more information, go to [Template Anatomy][1] in the AWS CloudFormation User Guide.
+    ///
+    /// [1]: https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/template-anatomy.html
+    Inline { body: String },
+
+    /// Location of file containing the template body.
+    ///
+    /// The URL must point to a template (max size: 460,800 bytes) that is located in an Amazon S3
+    /// bucket. For more information, go to the [Template Anatomy][1] in the AWS CloudFormation User
+    /// Guide.
+    ///
+    /// [1]: https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/template-anatomy.html
+    S3 { url: String },
+}
+
+impl TemplateSource {
+    /// Construct an [`Inline`](Self::Inline) template source for the given `body`.
+    #[must_use]
+    pub fn inline(body: impl Into<String>) -> Self {
+        Self::Inline { body: body.into() }
+    }
+
+    /// Construct an [`S3`](Self::S3) template source for the given `url`.
+    #[must_use]
+    pub fn s3(url: impl Into<String>) -> Self {
+        Self::S3 { url: url.into() }
+    }
+}
+
 /// The output of the `apply` operation.
 #[derive(Debug, PartialEq)]
 #[allow(clippy::module_name_repetitions)]
